/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   General file handling routines.

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

#include "gpcsym.h"
/* File Descriptor Record definition */
#include "gpc_fdr.h"
#include <sys/stat.h>

#ifdef USG
# include <string.h>
# include <fcntl.h>
#else
# include <strings.h>
# include <sys/file.h>
#endif

#define M_READ		1
#define M_WRITE		2
#define M_APPEND	4

static char *
strdup (p)
char *p;
{
    char *ret;
    if (!p)
	return NULL;
    if ((ret = malloc (strlen(p)+1)) == NULL)
	_p_error(ABORT, "Out of HEAP in strdup");
    strcpy(ret, p);
    return ret;
}

void
_p_initfdr(File, Name, Size, State)
FDR	 File;	/* address of FDR variable */
char    *Name;  /* address of internal name in program */
int	 Size;  /* file buffer size; in bits, if Packed, else in bytes */
int 	 State;	/* Lazy (8), Ext (4), Pack(2), Text(1) Flags */
{
    m_FILSTA(File) = 0;
    m_STATUS(File) = FiNOP;
    m_NXTFDR(File) = NULL;

    if (State & 1)
	set_TXT(File);
    if (State & 2)
	set_PCK(File);
    if (State & 4) {
	set_EXT(File);
	if (! Name)
	    _p_generic (703);
    }
    if (State & 8)
	set_LAZY (File);
    
    /* Name may be NULL for internal files */
    m_NAM (File) = Name;
    m_EXTNAM(File) = NULL;

    if (Size) {
	m_SIZ (File) = Size;
	if (tst_PCK (File)) {
	    /* Convert to unpacked file, since we don't yet support
	     * PACKED files. This may be done even if the compiler
	     * thinks we support them :-)
	     */
	    m_SIZ(File) = (m_SIZ(File) + 7) / 8;
	    clr_PCK(File);
	}
	/* Allocate file buffer from heap (don't use p_NEW :-) */
	if ((m_FILBPTR(File) = malloc (m_SIZ(File))) == NULL)
	    _p_error (ABORT, "No room for file buffers in heap");
    }

    /* Mark the file buffer contents undefined */
    set_UND(File);

    m_FILNUM(File) = NULL;
    D(1, _p_printfdr(File,"INITFDR"));
}

#ifdef CASECMP
/* Very slow strcasecmp(). Use the one in the C library. */
int
strcasecmp(s1, s2)
char *s1, *s2;
{
    char *S1 = strdup(s1), *S2 = strdup(s2);
    int res;
    
    s1 = S1;
    s2 = S2;
    if (s1)
	while(*s1) {
	    if (isupper(*s1))
		*s1 = tolower(*s1);
	    s1++;
	}

    if (s2)
	while(*s2) {
	    if (isupper(*s2))
		*s2 = tolower(*s2);
	    s2++;
	}
	
    res = strcmp(S1, S2);
    free(S1); free(S2);
    return res;
}
#endif

static char *
_p_nameit(File,mode)
FDR	File;
int	mode;
{
    int	tty;
    int	n, in, out;
    assoc	*ap;
    char	b[512];

    if (!tst_EXT(File)) {
	_p_makename(b);
	D(1, fprintf(stderr,"Opening internal file %s\n",b));
	return(strdup(b));
    }
    for(ap = Gpc_assoc; ap->int_name ; ap++) {
	if (strcasecmp(m_NAM(File),ap->int_name) == 0) {
	   ap->int_name = "?KukkuuRESET?"; /* Allow close(a); reset(a) to access next one */
	   m_EXTNAM(File) = strdup (ap->ext_name);
	   return(m_EXTNAM(File));
        }
    }
    if (tst_TXT(File) &&
	(strcmp(m_NAM(File),INPUT_FILE_NAME) == 0 ||
	 strcmp(m_NAM(File),OUTPUT_FILE_NAME) == 0))
	return(0);

    /* Try to write filename prompts to /dev/tty and try to read responces
       from there also, to avoid mungling with stdin & stdout.

       However, if all everything failes, try stdin & stdout, if they don't
       work, abort. Last resort is to use -a switch to associate internal names
       to external ones. It needs to be documented.

     */

    if ((tty = open ("/dev/tty", O_RDWR)) < 0) {
	_p_error (REPORT, "Failed to open terminal for file name read, using stdin & stdout");
	in  = 0;
	out = 1;
    } else 
	in = out = tty;

    sprintf(b,"%s file %s: ",
	    mode == M_READ ? "Input" : mode == M_WRITE ? "Output " : "Expanded",
	    m_NAM(File));
    if ((n = write (out,b,strlen(b))) != strlen(b)) {
	if (out != 1)
	    _p_error (REPORT, "Writing file name prompt to /dev/tty failed, using stdout");
	if (out == 1 || (n = write (1,b,strlen(b))) < 0)
	    _p_error (ABORT, "Can't prompt user for external file name bindings");
    }
    if ((n = read (in,b,sizeof(b))) < 0) {
	if (in) {
	    _p_error (REPORT, "Reading filename from /dev/tty failed, trying stdin");
	    /* B should be ok still, since read failed. */
	    (void) write (tty,b,strlen(b));
	}
	if ( ! in || (n = read (0,b,sizeof(b))) < 0)
	    _p_error (ABORT, "Can't query user for external file name bindings");
    }
    b[n] = '\0';
    if (n > 0 && b[n-1] == '\n') b[n-1] = '\0';
    close (tty);
    m_EXTNAM(File) = strdup(b);
    return m_EXTNAM(File);
}

static void
_p_fdrchain(Fdr, What)
FDR	Fdr;
int What;
{
    FDR scan = NULL;

    switch (What) {
    case TRUE:
	if (!FirstFdr) {
	    FirstFdr = LastFdr = Fdr;
	    m_NXTFDR(Fdr) = NULL;
	    D(4, fprintf(stderr, "Adding %s to empty fdr list\n",m_NAM(Fdr)));
	} else {
	    for(scan=FirstFdr; scan; scan = m_NXTFDR(scan)) {
		if (scan == Fdr) /* FDR already in list, do nothing */
		    break;
	    }
	    if (!scan) {
		D(4, fprintf(stderr, "Adding %s to fdr list\n", m_NAM(Fdr)));
		if (m_NXTFDR(LastFdr))
		    _p_error(REPORT, "LastFdr->nxtfdr is not NULL\n");
		m_NXTFDR(LastFdr) = Fdr;
		m_NXTFDR(Fdr) = NULL;
		LastFdr = Fdr;
	    }
	}
	break;
    case FALSE:
	if (FirstFdr == Fdr) {  /* First in the Fdr list */
	    FirstFdr = m_NXTFDR(Fdr);
	    m_NXTFDR(Fdr) = NULL;
	    if (LastFdr == Fdr) /* The only element */
		LastFdr = NULL;
	    D(4, fprintf(stderr, "Removed fdr %s (first) from list\n",
			 m_NAM(Fdr)));
	} else {
	    /* Look for the element before Fdr in the list */
	    for(scan=FirstFdr; m_NXTFDR(scan) != Fdr; scan = m_NXTFDR(scan))
		if (!m_NXTFDR(scan)) {
		    scan = NULL;
		    break;
		}
	    if (scan) {
		D(4, fprintf(stderr, "Removed fdr %s from list\n",m_NAM(Fdr)));
		m_NXTFDR(scan) = m_NXTFDR(Fdr); /* skip over */
		m_NXTFDR(Fdr)  = NULL;
		if (LastFdr == Fdr) LastFdr = scan;
	    } else
		_p_error(REPORT, "Fdr not found in list\n");
	}
	break;
    }
}

static int
_p_open(File, Filename, mode, length)
FDR	File;
char	*Filename;
int mode;
int length;
{
    char *errstr = (char *) NULL;
    char *filename = (char *) NULL;

    if (length) {
	filename = malloc (length+1);
	(void) strncpy (filename, Filename, length);
	filename[length] = '\000';
    }
    if (m_SIZ (File) == 0)
	_p_generic (701);	/* FDR not initialized */

    if (filename) {
	_p_close(File);
	m_EXTNAM(File) = filename;
    }

    if (m_STATUS(File) != FiNOP) {

	/* File is currently open in Pascal program */

	if (mode != M_READ && TST_STATUS (File, FiRONLY))
	    _p_error (ABORT, "Can't write to a read only file");
	
	if (mode == M_APPEND)
	    fseek(m_FILNUM(File),0L,2);		/* Start appending */
	else if (mode == M_READ)
	    fseek(m_FILNUM(File),0L,0);
	else {
#ifdef BSD
	    /* We have ftruncate () */
	    fseek(m_FILNUM(File),0L,0);
	    if (ftruncate(fileno(m_FILNUM(File)),0) == -1)
		_p_error (ABORT, "ftruncate failed when re-opening file with REWRITE");
#else
	    /* If you don't have ftruncate() emulate the behaviour */
	    filename = m_EXTNAM(File);
	    m_EXTNAM(File) = NULL;
	    _p_close(File);

	    m_EXTNAM(File) = filename;

	    /* Let the code below re-open the same external file for writing */
	    /* If the file is internal, it will not be the same, but who cares! */
#endif
	}
    }

    if (m_STATUS(File) == FiNOP) {
	while(m_FILNUM(File) == NULL) {
	    char *how;
	    if (mode == M_READ && !tst_EXT(File))
		_p_generic(13);
	    if (!filename)
		filename = _p_nameit(File, mode);
	    if (!filename)
		m_FILNUM(File) = (mode == M_READ ? stdin : stdout);
	    else if (! *filename)
		m_FILNUM(File) = NULL;
	    else {
		switch (mode) {
		case M_READ:
		    how = "r+";
		    errstr = "File can not be opened for reading";
		    break;
		case M_WRITE:
		    how = "w+";
		    errstr = "File can not be opened for writing";
		    break;
		case M_APPEND:
		    how = "a+";
		    errstr = "File can not be extended";
		    break;
		default:
		    abort ();
		}
		m_FILNUM(File) = fopen (filename, how);

		/* Now, if the file could not be opened, but we only want
		   to read it, check if it's possible
		*/
		if (mode == M_READ && m_FILNUM(File) == NULL) {
		    m_FILNUM(File) = fopen(filename,"r");
		    if (m_FILNUM(File) == NULL) {
			SET_STATUS (File, FiRONLY);
#ifdef WARN_READ_ONLY
			_p_warning ("File opened as a READ ONLY file");
#endif
		    }
		}
		
	    }
	    if (m_FILNUM(File) == NULL) {
		if (filename) {
		    (void) free (filename);
		    filename = (char *)NULL;
		}
		_p_error(REPORT, errstr);
	    } else {
		if (!tst_EXT(File)) {
		    extern char *Gpc_tmpname;
		    if (strncmp (filename,Gpc_tmpname, strlen(Gpc_tmpname)))
			_p_generic(700);
		    else if (!Gpc_debug)
			unlink(filename);
		}
	    }
	}
    }

    if (filename && ! m_EXTNAM(File))
	(void) free (filename);

    set_UND(File);	     /* Mark the file buffer contents undefined */
}

void
_p_rewrite (File, filename, length)
FDR File;
char *filename;
int length;
{
    struct stat finfo;

    _p_open (File, filename, M_WRITE, length);

    set_EMPTY(File);
    set_EOF(File);

    SET_STATUS(File, FiWRI);

    m_FISIZE(File) = 0;

    if (fstat(fileno(m_FILNUM(File)), &finfo) == 0) {
	if ((finfo.st_mode & S_IFMT) == S_IFCHR) {
#ifdef BSD
	    if (finfo.st_rdev == Gpc_DEVNULL)
		set_NUL(File);
	    else
#endif BSD
		if (isatty(fileno(m_FILNUM(File))))
		    set_TTY(File);
	}
    }
    _p_fdrchain(File, TRUE); /* Add to FDR chain */
    D(1, _p_printfdr(File,"REWRITE"));
}

void
_p_reset(File, filename, length)
FDR	 File;
char	*filename;
int	 length;
{
    struct stat finfo;

    _p_open(File, filename, M_READ, length);

    clr_EOF(File);
    clr_EOLN(File);

    SET_STATUS(File, FiORE);

    m_FISIZE(File) = 0;

    /* Please do not remove this. It's here as an inside joke */
    D(1, fprintf(stderr,"Kukkuu RESET (TM)\n"));

    if (fstat(fileno(m_FILNUM(File)), &finfo) == 0) {

	m_FISIZE(File) = NUMBYTE (File, finfo.st_size);

	if ((finfo.st_mode & S_IFMT) == S_IFCHR) {
#ifdef BSD
	    if (finfo.st_rdev == Gpc_DEVNULL)
		set_NUL(File);
	    else
#endif BSD
	    if (isatty(fileno(m_FILNUM(File))))
		set_TTY(File);
	}
    }
    if (m_FISIZE(File) == 0)
	set_EMPTY(File);
    else
	clr_EMPTY(File);

    _p_fdrchain(File, TRUE); /* Add FDR to chain if not there. */

    if (! tst_LAZY(File))
	if (tst_TXT(File) && tst_TTY(File)) {

	    /* This is implementation dependent! */
	    
	    D(2, fprintf(stderr, "Setting implicit EOLN to terminal\n"));
	    
	    m_FILBUF(File) = ' ';
	    set_EOLN(File);
	    clr_EMPTY(File);
	    clr_UND(File);
	} else
	    _p_get(File);	/* Read in the first buffer contents */
    
    D(1, _p_printfdr(File,"RESET"));
}

void
_p_extend(File, filename, length)
FDR   File;
char *filename;
int   length;
{
    struct stat finfo;

    _p_open (File, filename, M_APPEND, length);

    set_EOF(File);
    clr_EOLN(File);

    SET_STATUS(File, FiWRI);

    m_FISIZE(File) = 0;
    D(1, fprintf(stderr,"Appending to file\n"));
    if (fstat(fileno(m_FILNUM(File)), &finfo) == 0) {

	m_FISIZE(File) = BYTENUM (File, finfo.st_size);

	if ((finfo.st_mode & S_IFMT) == S_IFCHR) {
#ifdef BSD
	    if (finfo.st_rdev == Gpc_DEVNULL)
		m_STATUS(File) |= FiNUL;
	    else
#endif BSD
	    if (isatty(fileno(m_FILNUM(File))))
		m_STATUS(File) |= FiTTY;
	}
    }
    if (m_FISIZE(File) == 0)
	set_EMPTY(File);
    else
	clr_EMPTY(File);

    _p_fdrchain(File, TRUE); /* Add FDR to chain if not there. */
    D(1, _p_printfdr(File,"EXTEND"));
}

void
_p_close(File)
FDR	File;
{

    if (m_STATUS(File) == FiNOP) {
	return;
    }

    set_EOF(File);

    if (m_STATUS(File) & (FiANY)) {
	fclose(m_FILNUM(File));
	m_FILNUM(File) = NULL;
    }

    if (m_EXTNAM(File))
	free (m_EXTNAM(File));

    _p_fdrchain(File, FALSE); /* Remove from chain */

    _p_initfdr(File, m_NAM (File), NULL,
	       (tst_TXT(File) | tst_PCK(File) << 1 | tst_EXT(File) << 2 |
		tst_LAZY(File) << 3));

    D(1, _p_printfdr(File, "CLOSED"));
}

/* This can be inlined... */
int
_p_eof (File)
FDR	File;
{
    if (m_STATUS(File) == FiNOP)
	_p_generic(40);
    return tst_EOF (File);
}

/* This can also be inlined */
int
_p_eoln (File)
FDR	File;
{
    if (m_STATUS(File) == FiNOP)
	_p_generic(41);
    else if (tst_EOF(File))
	_p_generic(42);
	
    return tst_EOLN (File);
}

/* This is the buffer referencing routine, that the compiler
 * should do inline. Nothing is actually done, if tst_UND(File)
 * is not on. Compiler should dereference the file buffer address
 * to get the value from the buffer.
 */
void
_p_lazyget(File)
FDR File;
{
    /* If the file buffer contents is undefined, do a get */

    if (tst_UND(File))
	_p_get (File);
}

# ifdef DEBUG
_p_printfdr(File,s)
FDR	File;
char	*s;
{

    if (s) fprintf(stderr,"%s: ",s);
    fprintf(stderr,"FDR of   %.10s\n",m_NAM(File) ? m_NAM(File) :
	    "Internal");
    if (Gpc_debug >= 2) {
	fprintf(stderr,"\tofnum  %d\n", (m_FILNUM(File) ?
					 fileno(m_FILNUM(File)) :
					 -1));
	fprintf(stderr,"\tUND    %d\n",tst_UND(File));
	fprintf(stderr,"\tLAZY   %d\n",tst_LAZY(File));
	fprintf(stderr,"\tEOF    %d\n",tst_EOF(File));
	fprintf(stderr,"\tEOLN   %d\n",tst_EOLN(File));
	fprintf(stderr,"\tEXT    %d\n",tst_EXT(File));
	fprintf(stderr,"\tPCK    %d\n",tst_PCK(File));
	fprintf(stderr,"\tTXT    %d\n",tst_TXT(File));
	fprintf(stderr,"\tEMPTY  %d\n",tst_EMPTY(File));

	fprintf(stderr,"\tSTATUS 0x%x\n",m_STATUS(File));
	fprintf(stderr,"\tSIZ    %d\n",m_SIZ(File));
	fprintf(stderr,"\tExt name %s\n",m_EXTNAM(File) ? m_EXTNAM(File) : "<none>");
	fprintf(stderr,"\tElem   %d\n",m_FISIZE(File));  
	fprintf(stderr,"\tthis fdr adr: 0x%x\n", File);
	fprintf(stderr,"\tnext fdr adr: 0x%x\n",m_NXTFDR(File));
    }
}
# endif DEBUG
