/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   Random access file routines.

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

#include "gpcsym.h"
/* File Descriptor Record definition */
#include "gpc_fdr.h"
#include <sys/stat.h>

/* NOTE:
 *   Extended Pascal defined the following operations only to
 * DIRECT ACCESS FILE types:
 *
 *   SeekRead, SeekWrite, SeekUpdate, Empty, Position, LastPosition
 *   and Update (in not writing to end of file)
 *
 * Direct access file is defined by: FILE [ indextype ] OF type 
 * (the ord(a) in assertions means the smallest value of indextype)
 *
 * However, Gpc does not currently implement those files, and
 * anyhow maybe we should allow the operations also to other
 * files capable to seek(2). So, the way the files may currently be
 * viewed is
 *
 * TYPE Natural = 0..MaxInt;
 *      gpcfiles = FILE Natural OF <type>
 *
 * DefineSize is a GPC extension.
 *
 */

/*
 * Recalculate the m_FISIZE for file.
 * Take care you flush buffers properly!
 */
static void
_p_resize (File)
FDR File;
{
    struct stat finfo;

    if (fstat(fileno(m_FILNUM(File)), &finfo) == 0)
	m_FISIZE(File) = NUMBYTE (File, finfo.st_size);
}

/* DEFINESIZE (This is a GPC extension)
 *
 * May be applied only to random access files and files opened
 * for writing.
 *
 * Define files size as count of its component type units
 */
void
_p_definesize(File, NewSize)
FDR	File;
int  NewSize;				/* Should really be off_t */
{
    int ByteNum;

    if (m_STATUS(File) == FiNOP)
	_p_generic(407);
    else if (m_STATUS(File) & FiRONLY)
	_p_generic(411);
    else if (!(m_STATUS(File) & (FiRND | FiWRI)))
	_p_generic(404);

    if (NewSize < m_FISIZE(File)) {
	ByteNum = BYTENUM(File, NewSize);
	D(3, printf("DefineSize: trucating file to %d bytes\n", ByteNum));
#ifdef SYS5
	fprintf(stderr,"ftruncate needed by random_access\n");
	/* exit(); don't, it's not so fatal... */
#else BSD
	if (ftruncate(fileno(m_FILNUM(File)), ByteNum) == -1)
	    _p_error(ABORT, "DefineSize truncation failed");
#endif BSD
    }
    /* Just leave it like it is. */
    /* Of course, if the file is closed now, we are cheating. */
    /* Fix this later */
    m_FISIZE(File) = NewSize;
}

/* SEEKREAD
 * pre-assertion: (neither f0.L nor f0.R is undefined) and
 *		  ( 0<= ord(n)-ord(a) <= length(f0.L~f0.R))
 * post-assertion:(f.M = Inspection) and (f.L~f.R = f0.L~f0.R) and
 *		  (if length(f0.L~f0.R) > ord(n)-ord(a) then
 *		     ((length(f.L) =  ord(n)-ord(a)) and
 *		      (f^ = f.R.first))
 *		   else
 *		     ((f.R = S() and f^ is undefined)))
 */
void
_p_seekread(File, NewPlace)
FDR	File;
int NewPlace;	
{
    long ByteNum;

    if (m_STATUS(File) == FiNOP)
	_p_generic(407);
    else if (!(m_STATUS(File) & FiANY))
	_p_generic(409);
    else if (NewPlace < 0)
	_p_generic(410);
    else if (NewPlace > m_FISIZE(File))
	_p_generic(406);
    ByteNum = BYTENUM (File, NewPlace);

    D(3, printf("SeekRead: seek to %d\n", ByteNum));

    fflush (m_FILNUM(File));
    if (fseek(m_FILNUM(File), ByteNum, 0) == -1)
	_p_error(ABORT, "SeekRead seek failed");

    CLR_STATUS (File, FiANY);
    SET_STATUS (File, FiORE);

    clr_EOF (File);
    _p_get(File);

    /* Seek back to the place where we were before the
       GET. It's m_SIZ(File) before the place we are now at */

    fflush (m_FILNUM(File));
    if (fseek (m_FILNUM(File), -m_SIZ(File), 1))
	_p_generic(412);
}

/* SEEKWRITE
 * pre-assertion: (neither f0.L nor f0.R is undefined) and
 *		  ( 0<= ord(n)-ord(a) <= length(f0.L~f0.R))
 * post-assertion:(f.M = Generation) and (f.L~f.R = f0.L~f0.R) and
 *		  (length(f.L) = ord(n)-ord(a)) and (f^ is undefined)
 *
 * THIS DOES NOT WRITE ANYTHING ^^^^^^ ?????? I assume it is a bug
 * and write out the file buffer.
 *
 * FIX THIS when you have more information
 */
void
_p_seekwrite(File, NewPlace)
FDR	File;
int NewPlace;
{
    long ByteNum;

    if (m_STATUS(File) == FiNOP)
	_p_generic(407);
    else if (m_STATUS(File) & FiRONLY)
	_p_generic (411);
    else if (!(m_STATUS(File) & FiANY))
	_p_generic(409);
    else if (NewPlace < 0)
	_p_generic(410);
    else if (NewPlace > m_FISIZE(File)) {
	_p_warning(405);
	_p_definesize(File, NewPlace);
    }

    ByteNum = BYTENUM(File, NewPlace);

    D(3, printf("SeekWrite: seek to %d\n", ByteNum));
    fflush (m_FILNUM(File));
    if (fseek(m_FILNUM(File), ByteNum, 0) == -1)
	_p_error(ABORT, "SeekWrite seek failed");

    CLR_STATUS (File, FiANY);
    SET_STATUS (File, FiWRI);

    _p_put(File);

    /* Seek back to the place where we were before the
       PUT. It's m_SIZ(File) before the place we are now at */

    fflush (m_FILNUM(File));
    if (fseek (m_FILNUM(File), -m_SIZ(File), 1))
	_p_generic(412);

    _p_resize (File);
}

/* SEEKUPDATE
 * pre-assertion: (neither f0.L nor f0.R is undefined) and
 *		  ( 0<= ord(n)-ord(a) <= length(f0.L~f0.R))
 * post-assertion:(f.M = Update) and (f.L~f.R = f0.L~f0.R) and
 *		  (if length(f0.L~f0.R) > ord(n)-ord(a) then
 *		     ((length(f.L) =  ord(n)-ord(a)) and
 *		      (f^ = f.R.first))
 *		   else
 *		     ((f.R = S()) and (f^ is undefined)))
 *
 * The difference with SEEKREAD is that this leaves f.M to UPDATE mode.
 */
void
_p_seekupdate(File, NewPlace)
FDR File;
int NewPlace;
{
    long ByteNum;

    if (m_STATUS(File) == FiNOP)
	_p_generic(407);
    else if (m_STATUS(File) & FiRONLY)
	_p_generic (411);
    else if (!(m_STATUS(File) & FiANY))
	_p_generic(409);
    else if (NewPlace < 0)
	_p_generic(410);
    else if (NewPlace > m_FISIZE(File))
	_p_generic(406);

    ByteNum = BYTENUM(File, NewPlace);

    D(3, printf("SeekUpdate: seek to %d\n", ByteNum));
    fflush (m_FILNUM(File));
    if (fseek(m_FILNUM(File), ByteNum, 0) == -1)
	_p_error(ABORT, "SeekUpdate seek failed");

    CLR_STATUS (File, FiANY);
    SET_STATUS (File, FiRND);

    clr_EOF(File);
    _p_get(File);
    
    /* Seek back to the place where we were before the
       GET. It's m_SIZ(File) before the place we are now at */

    fflush (m_FILNUM(File));
    if (fseek (m_FILNUM(File), -m_SIZ(File), 1))
	_p_generic(412);
}

/* EMPTY 
 *
 * Returns True if file is Empty, otherwise False
 * Should require a direct access file type.
 */
char			/* Or whatever BOOLEAN is */
_p_empty (File)
FDR File;
{
    if (! (m_STATUS(File) & FiANY))
	_p_generic(407);

    return tst_EMPTY(File);
}

/* UPDATE
 * pre-assertion: (f0.M = Generation or f0.M = Update) and
 *		  (neither f0.L nor f0.R is undefined) and
 *		  (f0.R = S() or f is a direct access file type) and
 *		  (f0^ is not undefined)
 * post-assertion:(f.M = f0.M) and (f.L = f0.L) and
 *		  (if f0.R = S() then
 *		     (f.R = S(f0^))
 *		   else
 *		     (f.R = S(f0^)~f0.R.rest)) and
 *		   (f^ = f0^).
 *
 * i.e. Write the stuff in, and leave it also in the file buffer.
 * don't advance the file pointer from the pre-assert state!
 */
/* THIS DOES NOT WORK. Should not move file pointer!!! */
void
_p_update (File)
FDR File;
{

    /* Trust that PUT does not clobber f0^
       If it does clobber, you need to save it before
       calling and restore after return. */
    _p_put(File);

    /* Seek back to the place where we were before the
       PUT. It's m_SIZ(File) before the place we are now at */

    fflush (m_FILNUM(File));
    if (fseek (m_FILNUM(File), -m_SIZ(File), 1))
	_p_generic(412);
}

/* LASTPOSITION
 * LastPosition(f) = succ(a, length(f.L~f.R)-1);
 */
int
_p_lastposition(File)
FDR File;
{

    if (m_STATUS(File) == FiNOP)
	_p_generic(407);
#ifdef ALLOW_RND_ONLY
    else if (!(m_STATUS(File) & FiRND))
	_p_generic(400);
#endif
    _p_resize (File);

    return(m_FISIZE(File));
}

/* POSITION
 * Position(f) = succ(a, length(f.L));
 *
 * Should really return a LONG!!!
 */
int
_p_position(File)
FDR File;
{
    long NumBytes;

    if (m_STATUS(File) == FiNOP)
	_p_generic(407);
#ifdef ALLOW_RND_ONLY
    else if (!(m_STATUS(File) & FiRND))
	_p_generic(400);
#endif

    NumBytes = ftell (m_FILNUM (File));

    return NUMBYTE (File, NumBytes);
}
