/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   Routines to read various things from files. There are two entry points
   that the compiler should call: _p_get and _p_read.

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

/*
 * Author: Jukka Virtanen <jtv@hut.fi>
 */

#include "gpcsym.h"
/* File Descriptor Record definition */
#include "gpc_fdr.h"
#include "gvarargs.h"

# define ROK     (0)
# define RSYNTAX (-1)
# define RFATAL  (-2)

#define NO_SCANF	/* SCANF does not trap integer overflows
			   when reading numbers. (It's not a bug,
			   actually. Just shifting).
			    It does not scanf the floating point numbers
			   as the standard requires. So, do the syntax
			   analysis here and give the result to
			   sscanf. */

static void _p_GetByt();

#ifndef NO_SCANF

static int
_p_readit(File,format,p)
FDR	File;
char	*format;
char	*p;
{
    FILE *f;
	
    /* Push back the character we read before */

    /* Since leading spaces do not matter when reading numbers,
       don't push the possible leading space back */

    /* Be sure you have read it before entry
       or set it to space (like implicit EOLN :-) */

    f = m_FILNUM(File);

    if (! isspace(m_FILBUF(File)) && ungetc(m_FILBUF(File),f) == EOF) {
	_p_error(ABORT, "ungetc");
	return(RFATAL);
    }
    
    if (fscanf(f,format,p) != 1) {
	return(RSYNTAX);
    }
    return(ROK);
}
/* end of FSCANF using code */

#else

/* Use this stuff to read integers, if your fscanf does not trap
   overflows or it can not parse floating point numbers. */

static int
getint(Fil, ch)
FILE *Fil;
register int ch;
{
    register int num = 0;

    /* ch is a digit on entry */

#ifdef ENABLE_TRAPS
    /* This thing enables traps while reading in numbers */
    ENABLE_TRAPS;
#endif
    Gpc_sigcause = 606;	     /* Overflow while reading integer */
    Gpc_signal   = SIGFPE;   /*  if SIGFPE occurs now */
    do {
	num = num*10+(ch-'0'); /* To trap overflows, don't shift */
	ch  = getc(Fil);
    } while (isdigit(ch));
    Gpc_signal = 0;
    ungetc(ch, Fil);	/* Push back last non-digit */
    return(num);
}

static int
_p_rdint(File,p)
FDR	File;
int	*p;
{
    register FILE *f;
    int negative, num;
    int ch;

    /* file buffer is JUNK after this till return to _p_read */

    f = m_FILNUM(File);

    ch = m_FILBUF(File);	       /* Get the buffered chraco */

    negative = FALSE;
    while(isspace(ch))
	ch = getc(f);

    if (!isdigit(ch) && ch != '+' && ch != '-') {
	_p_generic(605); /* Sign or digit expected */
	ch = '0';
    } else {
	if (ch=='+' || ch=='-') {
	    if (ch == '-')
		negative = TRUE;
	    ch = getc(f);
	    if (!isdigit(ch)) {
		_p_generic(604); /* Digit expected after sign */
		ch = '0';       /* If returned, set it to something */
	    }
	}
    }
    
    /* Now the 'ch' contains the first digit */
    num = getint(f, ch);
    *p  = (negative? -num : num);

    return(ROK);
}

/*
 * SCANF can not parse real numbers according to Pascal standard.
 * Do it yourself, again.
 */
static char *where;

static void
store_int (what)
     int what;
{
  if (what > 9)
    store_int (what/10);

  *where++ = (what%10)+'0';
}


static int
_p_rdfloat(File, res)
FDR	File;
register char *res; /* stores characters in here */
{
    register FILE *f;
    register int  ch;
    char *mantissa_start;
    int nonzero    = 0;
    int adjust_exp = 0;
    int expon = 0;
    int esign = 1;

    /* file buffer is JUNK after this till return to _p_read */

    f = m_FILNUM(File);

    ch = m_FILBUF(File);	       /* Get the buffered chraco */

    while(isspace(ch))
	ch = getc(f);
    if (!isdigit(ch) && ch != '+' && ch != '-') {
	_p_generic(605); /* Sign or digit expected */
	*res = '0';
	return(ROK);
    } else {
	if (ch=='+' || ch=='-') {
	    *res++ = ch;
	    ch = getc(f);
	    if (!isdigit(ch)) {
		_p_generic(604); /* Digit expected after sign */
		*res = '0';
		return(ROK);
	    }
	}
    }

    mantissa_start = res;

    /* Read the mantissa */
    do {
      if (ch != '0' || nonzero)
	{
	  *res++ = ch;
	  nonzero = 1;
	}
      ch = getc(f);
    } while(isdigit(ch));

    /* read the fractional part */
    if (ch == '.')  { /* Read the fractional part */
      if (nonzero)
	*res++ = ch;	/* Store decimal point if mantissa nonzero */
      else
	adjust_exp = -1;

      ch = getc(f);
      if (!isdigit(ch)) {
	_p_generic(607); /* Digit expected after decimal point */
	ch = '0';
      }

      do {
	if (nonzero || ch != '0')
	  {
	    *res++ = ch;
	    if (! nonzero)
	      {
		*res++ = '.';
		nonzero = 2;
	      }
	  }
	else
	  adjust_exp -= 1;

	ch = getc(f);
      } while(isdigit(ch));
      if (!nonzero)
	{
	  *res++ = '0';
	  *res++ = '.';
	  *res++ = '0';
	  adjust_exp = 0;
	}
    }

    /* read the exponent */
    if (ch=='e' || ch=='E') {

      if (nonzero)
	*res++ = ch;

      ch = getc(f);
      if (ch == '+' || ch == '-') {
	if (nonzero)
	  *res++ = ch;
	if (ch == '-')
	  esign = -1;
	ch = getc(f);
      }
      if (!isdigit(ch)) {
	_p_generic(608); /* Digit expected while reading exponent */
	ch = '0';
      }
      do {
	expon = 10*expon+(ch-'0');
	ch = getc(f);
      } while(isdigit(ch));
    }

    if (nonzero && (adjust_exp || expon))
      {
	if (expon)
	  /* Add the adjust value */
	  expon = esign*expon+adjust_exp;
	else
	  {
	    expon = adjust_exp;
	    *res++ = 'E';
	    if (expon < 0)
	      {
		esign  = -1;
		*res++ = '-';
	      }
	  }

	if (expon > DBL_MAX_10_EXP || expon < DBL_MIN_10_EXP)
	  _p_generic(609);

	expon = esign*expon;
	where = res;
	store_int (expon);
	res = where;
      }

    *res++ = '\000';
    ungetc(ch, f); /* Push back the char not belonging to the number */

    return(ROK);
}
#endif NO_SCANF

static void
_p_readi(File, ret)
FDR	File;
int	*ret;
{

#ifdef NO_SCANF
    switch (_p_rdint (File, ret)) 
#else
    switch (_p_readit (File,"%d", ret)) 
#endif
    {
    case ROK:
	break;
    case RSYNTAX:
	_p_generic(600);
	*ret = 0;
	break;
    }
}

static void
_p_readr(File, real)
FDR	File;
double  *real;
{
#ifdef NO_SCANF
    char buffer[ 256 ];
#endif
	
#ifdef NO_SCANF
    switch(_p_rdfloat(File, buffer))
#else
    switch (_p_readit(File,"%lf", real)) 	/* long float == double */
#endif
    {
    case ROK:
#ifdef NO_SCANF
	/* Well, use it to convert to binary */
	sscanf(buffer, "%lf", real);
#endif
	break;
    case RSYNTAX:
	_p_generic(601);
	*real = 0.0;
	break;
    }
}


/* Routine reads various things from TEXT files by
 * calling the subroutines above to do the actual job.
 */

/* Pre-assertion:  same as GET below
 * Post-assertion: (f.M = f1.M) and
 *		   (f.L~f.R = f0.L~f0.R) and
 *		   (f.R = t~u) and
 *		   (f^ = f.R.first)
 */

void
_p_read (va_alist)
va_dcl
{
    register va_list p;
    FDR File;

    int length;
    int failed = 0;

    va_start (p);
    File   = va_arg (p, FDR);	/* First arg is the file we read from */
    length = va_arg (p, int);	/* Second is the number of remaining codes */

    /* If you have non Pascal subroutines that output to terminal,
       but don't fflush, you might want to define READ_FLUSH in the
       Makefile to see what they write before you read from tty.

       It affects only the way how LAZY I/O flushes terminal output.
     */

#ifdef READ_FLUSH
    if (tst_TTY(File))
#else
    if (tst_TTY(File) && !tst_LAZY(File))
#endif
	_p_fflush(TRUE);

    if (!ok_READ(File))
	failed = 1;

    for (length; length > 0; length--) {
	int code = va_arg (p, int);	/* Type we are reading */

	if (tst_UND(File))
	    _p_GetByt (File);

	/* File buffer is now valid or EOF is set */

	switch (code) {
	case P_INT:
	    { int *i = va_arg (p, int *);
	      if (failed || !ok_EOF(File))
		  *i = 0;
	      else
		  _p_readi(File, i);
	      break;
	    }
	case P_CHAR:
	    { char *ch = va_arg (p, char *);
	      if (failed || !ok_EOF(File))
		  *ch = ' ';
	      else
		  *ch = m_FILBUF(File);
	      break;
	    }
	case P_REAL:
	    { double *d = va_arg (p, double *);
	      if (failed || !ok_EOF(File))
		  *d = 0.0;
	      else
		  _p_readr (File, d);
	      break;
	    }
	case P_LINE:
	    if (length != 1)
		_p_error (ABORT, "Compiler calls READLN incorrectly");
	    if (failed || !ok_EOF(File)) 
		break;

	    if (!tst_EOLN(File)) {
		_p_GetByt(File);
		while(!tst_EOLN(File))
		    _p_GetByt(File);
	    }
	    break;
#if 0
	    /* String type layout not decided yet, so the parameters of this
	       are not known yet. */
	case P_STRING:
	    { char *str = va_arg (p, char *); /* struct with length? */
	      int  maxlen = va_arg (p, int);
	      if (failed || !ok_EOF(File))
		  *str = '\000';
	      else
		  _p_reads (File, str, maxlen);
	      break;
	    }
#endif
	default:
	    _p_error (ABORT, "unknown code in _p_read");
	}

	/* If this is not a lazy file, validate the buffer contents */

	if (! tst_LAZY(File))
	    _p_GetByt(File);
	else
	    set_UND(File);
    }
    va_end (p);

}

/* Gets ONE BYTE from the file. */
static void
_p_GetByt(File)
FDR	File;
{
    if (ok_EOF(File)) {
	int temp;

	if (tst_TTY(File))
	    _p_fflush(TRUE);

	clr_UND(File);
	m_FILBUF(File) = temp = getc(m_FILNUM(File));
	if (temp == EOF) {
	    if (tst_TXT(File) && !tst_EOLN(File) && !tst_EMPTY(File)) {

		/* When reading from TEXT file EOLN is always true
		   just before EOF, even if there is no end of line
		   at the end of the file */

		set_EOLN(File);
		m_FILBUF(File) = ' ';
	    } else {
		set_EOF(File);
		clr_EOLN(File);
	    }
	    return;
	}
    
	if (tst_TXT(File) && m_FILBUF(File) == '\n') {
	    set_EOLN(File);
	    m_FILBUF(File) = ' ';
	} else
	    clr_EOLN(File);
    }
}

/* Gets m_SIZ bytes (> 1) from the file. */
static void
_p_GetN(File)
FDR	File;
{
    int n;

    if (ok_EOF(File)) {
	if ((n = fread(m_FILBPTR(File),1,m_SIZ(File),m_FILNUM(File))) < m_SIZ(File)) {
	    if (n != 0) {
		_p_error(REPORT, "GET partial record");
	    } else {
		set_EOF(File);
		clr_EOLN(File);
	    }
	}
	clr_UND(File);
    }
}

/* GET
 * pre-assertion: (f0.M = Inspection or f0.M = Update) and
 *		  (neither f0.L nor f0.R is undefined) and
 *		  (f0.R <> S())
 * post-assertion:(f.M = f0.M) and (f.L = f0.L~S(f0.R.first)) and
 *		  (f.R = f0.R.rest)) and
 * 		  (if (f.R = S()) then
 *		      (f^ is undefined)
 *		    else
 *		      (f^ = f.R.first))
 *
 */
void
_p_get(File)
FDR	File;
{
    if (ok_READ(File)) {
	if (m_SIZ(File) == 1) /* No files are packed yet. */
	    _p_GetByt(File);
	else
	    _p_GetN(File);
    }
}

