/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   Operations for SET type.

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

/*
 * Author: Jukka Virtanen <jtv@hut.fi>
 *
 */

#include "gpcsym.h"

#include "gpc_set.h"

/* Code clears the setA bits outside bounds.
 */
static void
rts_clear_outside (setA, lo, hi)
set setA;
int lo;
int hi;
{
  if (lo >= LO (setA) && hi <= HI (setA) && lo <= hi)
    {
      long nwl = THE_WORD (lo, setA);
      long nwh = THE_WORD (hi, setA);
      int  nbl = THE_BIT (lo);
      int  nbh = THE_BIT (hi);

      long owl = THE_WORD (LO (setA),  setA);
      long owh = THE_WORD (HI (setA), setA);
	
      if (owl == owh)			/* Is it the only word we have? */
	*(BITV(setA)+nwl) &= (~(-2 << (nbh - nbl))) << nbl;
      else
	{
	  int count = nwl - owl;
	  
	  while (count > 0)	/* Clear leading words */
	    BITV(setA)[ count-- ] = 0;
	  
	  if (nwl == nwh)
	    *(BITV(setA)+nwl) &= (~(-2 << (nbh - nbl))) << nbl;
	  else
	    {
	      /* Bits to leave on in the first word */
	      *(BITV(setA)+nwl) &= -1 << nbl;
	      
	      /* The words between are OK */
	      
	      /* Bits to leave on in the last word */
	      *(BITV(setA)+nwh) &= ~(-2 << nbh);
	    }

	  /* Clear trailing words */
	  count = owh - nwh;
	  while (count > 0)
	    BITV(setA)[ count-- ] = 0;
	}
    }
}

/* Internal routine to count elements of the set */
static long
rts_card (setA)
set setA;
{
  long wordsA   = NUM_WORDS (setA);
  long members  = 0;
  SET_VECTOR *a = BITV (setA);

  if (HI(setA) == -1 && LO(setA) == 0)
    return 0;

  rts_clear_outside (setA, LO(setA), HI(setA));

  while (wordsA--) {
    if (*a)
      if (*a == -1)
	members += BITS_PER_WORD;
      else {
	SET_VECTOR word = *a;
	while (word) {
	  members += !! (word & (1 << (BITS_PER_WORD-1)));
	  word <<= 1;
	}
      }
    a++;
  }
  return members;
}

#ifdef DEBUG
Dumpset (setA, level, op)
set setA;
int level;
char *op;
{
  int low	= LO (setA);
  int high 	= HI (setA);
  int words 	= NUM_WORDS (setA);
  int elements  = SET_SIZE (setA);
  int align_low = RLO(setA);
  int is_sparse = SPARSE_SET (setA);
  int members;
  int is_empty;

  int old_debug = Gpc_debug;
  
  if (Gpc_debug)
    {
      if (Gpc_debug > level)
	level = Gpc_debug;
  
      Gpc_debug  = 0;
      
      members = rts_card (setA);
      is_empty = EMPTY_SET (setA);
      
      printf ("%s: Set limits: `%d'..`%d' low element align %d\n",
	      op, low, high, align_low);
      printf ("%s: %s set; size %d elements; %d words; card %d\n",
	      op, is_empty ? "Empty" : is_sparse ? "Sparse" : "Normal",
	      elements, words, members);
      if (level > 1 && !is_empty && !is_sparse)
	{
	  int i;
	  SET_VECTOR *vector = BITV (setA);
	  printf ("Setd element bits (bit 0 on the RIGHT)\n");
	  printf ("setd: ");
	  for (i = 0; i < words; i++)
	    {
	      if (i && (i % 8 == 0))
		{
		  printf ("\nsetd: ");
		}
	      printf ("%08x ", *(vector+words-i-1));
	    }
	  printf ("\n\n");
	}
      Gpc_debug = old_debug;
    }
}
#endif

/*
 * Sets are aligned on word boundaries, first element is in the
 * first word of the set, bit THE_BIT(LO(setA)).
 * The bit number of the first element on word 0 is RLO(set), which  does
 * not belong to the set if felem%BITS_PER_WORD != 0
 *
 * The last bit HI(setA) is in the last word of the set,
 * in bit THE_BIT(HI(setA)).
 *
 * There may be unused bits in the end of the set, due to alignment.
 *
 * The number of words in the set is NUM_WORDS(set)
 *
 */

/* IN:
 *	setA, elem
 *	
 * OUT:
 * 	function returns 1 if ELEM is a member of set A else returns 0
 */
int
_p_set_in (setA, elem)
set setA;
long elem;
{
  D(1, Dumpset(setA, 2, "in"));

  if (EMPTY_SET(setA) || LO (setA) > elem || elem > HI (setA))
    return 0;

  return !!(BITV (setA)[ THE_WORD (elem, setA) ] & (1 << THE_BIT (elem)));
}

/*
 * R = A
 *
 * Assign (copy) A to R
 */
SET_VECTOR *
_p_set_copy (setA, setR)
set setA;
set setR;
{
  int hi = HI(setA);
  int lo = LO(setA);
  D(1, Dumpset(setA, 2, "copy (source)"));
  if (EMPTY_SET (setA))
    bzero (BITV (setR), NUM_BYTES (setR));
  else
    {
      int  wordsR = NUM_WORDS (setR);
      int  wordsA = NUM_WORDS (setA);

      if (wordsA <= wordsR && LO(setA) == LO(setR))
	{
	  bcopy (BITV (setA), BITV (setR), NUM_BYTES(setA));
	}
      else
	{
	  int first_word;

	  if (HI(setA) < LO(setR) || LO(setA) > HI(setA))
	    _p_error (ABORT,"Sets have no common element in _p_set_copy");

	  /* note: This copies only PARTIAL SET */
	  if (LO(setA) <= LO(setR) && HI(setA) >= LO(setR))
	    LO(setA) = LO(setR);

	  if (HI(setA) >= HI(setR) && LO(setA) <= HI(setR))
	    HI(setA) = HI(setR);

	  first_word = THE_WORD (LO(setA), setR);	  

	  bcopy (BITV(setA), &(BITV (setR)[ first_word+1 ]), NUM_BYTES (setA));
	}

      /* Clear out the bits not in range */
      rts_clear_outside (setR, LO(setR), HI(setR));
    }
  D(1, Dumpset(setR, 2, "copy (result)"));
  return BITV(setR);
}

/* TRAP ON BITMASK (runtime check)
 *
 * Return 1 if setA contains 1's in the bits that setB contains 0's
 * else return 0.
 *
 */
int
_p_set_trap (setA, setB)
set setA;
set setB;
{
  int wordsA = NUM_WORDS (setA);
  int wordsB = NUM_WORDS (setB);

  SET_VECTOR *a, *b;

  D(1, Dumpset(setA, 2, "trap (a)"));
  D(1, Dumpset(setB, 2, "trap (b)"));

  if (EMPTY_SET (setA))
    return 0;

  if (SPARSE_SET (setA) || SPARSE_SET (setB))
    abort ();

  if (wordsA != wordsB || LO(setA) != LO(setB))
    _p_error (ABORT, "compiler has generated an illegal set trap mask");

  a = BITV (setA);
  b = BITV (setB);

  while (wordsA--) {
    if (*a++ & ~*b++)
      return 1;
  }
  return 0;
}


/*
 * A = A + [ R_LOW..R_HIGH ]
 *
 * Mark each member from R_LOW to R_HIGH as being in the set
 *
 * If R_LOW  > R_HIGH do nothing (give a warning)
 * If R_LOW == R_HIGH it's not a range, it's a single member, so act
 * 		      accordingly.
 * 
 * returns a pointer to the resulting set
 *
 * SETR     is the resulting set
 * COUNT    is the number of range pairs
 * va_alist variable number of int pairs
 */
SET_VECTOR *
_p_set_range (va_alist)
va_dcl
{
  register va_list p;

  set setR;
  int elem_size;
  int count;
  long r_low;
  long r_high;

  long low;
  long high;
  long mask;

  va_start (p);
  setR  = va_arg (p, set);
  elem_size = va_arg (p, int);
  count = va_arg (p, int);
  
  switch (elem_size) {
    case 1: mask = 0xff;       break;
    case 2: mask = 0xffff;     break;
    case 4: mask = 0xffffffff; break;
#ifdef __alpha
    case 8: mask = 0xffffffffffffffff; break;  /* :-) */
#endif
    default:
      _p_error (705);
  }

  high = HI (setR);
  low  = LO (setR);
  while (count-- > 0)
    {
      /* @@ Maybe incorrect to assume that INT is the smallest
       * unit pushed to stack.
       */
      r_low  = elem_size == 8 ? va_arg (p, long) : va_arg (p, int);
      r_high = elem_size == 8 ? va_arg (p, long) : va_arg (p, int);

      r_low  &= mask;
      r_high &= mask;
      
      if (r_low > r_high) {
	_p_warning (509);
	return BITV(setR);
      }
      
      /* @@@ Maybe this is a fatal error after all. Should look this up from
	 the manual when I locate it again */
      if (r_low < low) {
	_p_warning (50);
	r_low = low;
      }

      /* @@@ ditto */
      if (r_high > high) {
	_p_warning (50); /* Gee, maybe I should get rid of these stupid numbers! */
	r_high = high;
      }

      if (r_low == r_high) { /* Set a single member */
	BITV (setR)[ THE_WORD(r_low, setR) ] |= 1 << THE_BIT(r_low);
	continue;
      }

      /* Now we need to set a valid range of bits between r_low & r_high */
      {
	long wl = THE_WORD (r_low,  setR);
	long wh = THE_WORD (r_high, setR);
	int  bl = THE_BIT (r_low);
	int  bh = THE_BIT (r_high);
	
	SET_VECTOR *setw = BITV (setR)+wl; /* Start modifying at this word */

	if (wl == wh)			/* Is it the only word we modify? */
	  *setw |= (~(-2 << (bh - bl))) << bl; /* Yes, so we are done */
	else {
	  *setw++ |= -1 << bl;		/* Bits to set in the first word */
	  
	  while (wl++ < wh)			/* The words between are all ones */
	    *setw++ = -1;
	  
	  *setw |= ~(-2 << bh);		/* Bits to set in the last word */
	}
      }
    }
  va_end (p);

  D(1, Dumpset(setR, 2, "range (result)"));
  return BITV(setR);
}

/*
 * A = A + [ element ]
 *
 * Mark one member of the set as being in it.
 *
 * returns a pointer to the resulting set
 *
 */
SET_VECTOR *
_p_set_single (va_alist)
va_dcl
{
  register va_list p;

  set  setR;
  int  elem_size;
  int  count;
  long elem;

  int  low;
  int  high;
  long mask;

  va_start (p);
  setR  = va_arg (p, set);
  elem_size = va_arg (p, int);
  count = va_arg (p, int);
  
  switch (elem_size) {
    case 1: mask = 0xff;       break;
    case 2: mask = 0xffff;     break;
    case 4: mask = 0xffffffff; break;
#ifdef __alpha
    case 8: mask = 0xffffffffffffffff; break;  /* :-) */
#endif
    default:
      _p_error (705);
  }

  high = HI (setR);
  low  = LO (setR);
  while (count-- > 0)
    {
      /* @@ Maybe incorrect to assume that INT is the smallest
       * unit pushed to stack.
       */
      elem  = elem_size == 8 ? va_arg (p, long) : va_arg (p, int);
      elem &= mask;
      
      if (elem < low || high < elem)
	{
	  D(1, fprintf (stderr, "When adding single %d into (%d .. %d)\n", elem, low,high));
	  _p_generic (50);
	}
      else
	BITV (setR)[ THE_WORD(elem, setR) ] |= (1 << THE_BIT(elem));
    }
  va_end (p);
  
  D(1, Dumpset(setR, 2, "single (result)"));
  return BITV(setR);
}

/*
 * R = A + B (UNION)
 * 
 * Returns a pointer to the resulting set R.
 */

SET_VECTOR *
_p_set_union (setA, setB, setR)
set setA;
set setB;
set setR;
{
  D(1, Dumpset(setA, 2, "union + (a)"));
  D(1, Dumpset(setB, 2, "union + (b)"));

  /* First, try the easy cases */

  if (EMPTY_SET (setA) && EMPTY_SET (setB))
    bzero (BITV (setR), NUM_BYTES (setR));
  else if (EMPTY_SET (setB))
    _p_set_copy (setA, setR);
  else if (EMPTY_SET (setA))
    _p_set_copy (setB, setR);
  else
    {
      int wordsB;
      int wordsR;
      
      SET_VECTOR *b = BITV (setB),
      		 *r = BITV (setR);
      
       _p_set_copy (setA, setR);

      if (LO(setB) < LO(setR) || HI(setB) > HI(setR) ||
	  LO(setB) > HI(setR) || HI(setB) < LO(setR))
	_p_error (ABORT, "_p_set_union internal error");
      
      if (LO (setB) > LO (setR))
	{
	  b += VECTOR_ADJUST (setB, setR);
	  LO(setB) = LO(setR);
	}

      wordsB = NUM_WORDS (setB);
      
      if (LO(setB) != LO(setR))
	r += THE_WORD (LO(setB), setR);
	  
      while (wordsB--)
	{
	  *r++ |= *b++;
	}
    }
  D(1, Dumpset(setR, 2, "union (results)"));
  return BITV(setR);
}

/*
 * R = A - B (DIFFERENCE)
 * 
 * Returns a pointer to the resulting set R.
 */

SET_VECTOR  *
_p_set_diff (setA, setB, setR)
set setA;
set setB;
set setR;
{
  D(1, Dumpset(setA, 2, "diff - (a)"));
  D(1, Dumpset(setB, 2, "diff - (b)"));
  /* First, try the easy cases */
  if (EMPTY_SET (setB))
    _p_set_copy (setA, setR);
  else if (EMPTY_SET (setA))
    bzero (BITV (setR), NUM_BYTES (setR));
  else
    {
      _p_set_copy (setA, setR);

      if (LO(setB) <= HI(setR) && HI (setB) >= LO (setR))
	{
	  int wordsB;

	  SET_VECTOR *b = BITV (setB),
	  	     *r = BITV (setR);
      
	  if (LO(setB) < LO(setR))
	    {
	      b += VECTOR_ADJUST (setB, setR);
	      LO(setB) = LO(setR);
	    }

	  if (HI(setB) > HI(setR))
	    HI(setB) = HI(setR);

	  wordsB = NUM_WORDS (setB);

	  if (LO(setB) != LO(setR))
	    r += THE_WORD (LO(setB), setR);
	  
	  while (wordsB--)
	    {
	      *r++ &= ~*b++;
	    }
	  rts_clear_outside (setR, LO(setA), HI(setA));
	}
    }
  D(1, Dumpset(setR, 2, "diff (result)"));;
  return BITV(setR);
}

/*
 * R = A * B (INTERSECTION)
 * 
 * Returns a pointer to the resulting set R.
 */

SET_VECTOR *
_p_set_intersection (setA, setB, setR)
set setA;
set setB;
set setR;
{
  D(1, Dumpset(setA, 2, "intersect * (a)"));
  D(1, Dumpset(setB, 2, "intersect * (b)"));
  /* First, try the easy cases */
  if (EMPTY_SET (setB) || EMPTY_SET (setA) ||
      LO(setA) > HI(setB) || HI(setA) < LO (setB))
    bzero (BITV (setR), NUM_BYTES (setR));
  else
    {
      /* Nonempty sets have at least some possible common elements */

      int wordsB;
      
      SET_VECTOR *b = BITV (setB),
      		 *r = BITV (setR);
      
      _p_set_copy (setA, setR);
      
      if (LO(setB) < LO(setR))
	{
	  b += VECTOR_ADJUST (setB, setR);
	  LO(setB) = LO(setR);
	}
      else if (LO(setB) > HI(setR))
	_p_error (ABORT, "set intersection internal error");
      else
	{
	  /* Do NOT adjust bounds here */
	  r += THE_WORD (LO(setB), setR);
	}

      if (HI(setB) > HI(setR))
	HI(setB) = HI(setR);
      
      /* Number of common words */
      wordsB = NUM_WORDS (setB);

      while (wordsB--)
	{
	  *r++ &= *b++;
	}

      /* Clear parts that don't overlap */
      rts_clear_outside (setR, LO(setB), HI(setB));
    }
  D(1, Dumpset(setR, 2, "intersection (result)"));
  return BITV(setR);
}

/*
 * A = B (EQUALITY)
 * 
 * Returns 1 if sets are equal, 0 otherwise.
 */

int
_p_set_equal (setA, setB)
set setA;
set setB;
{
  D(1, Dumpset(setA, 2, "equal (a)"));
  D(1, Dumpset(setB, 2, "equal (b)"));
  /* First, try the easy cases */
  if (EMPTY_SET (setB) && EMPTY_SET (setA))
    return 1;
  else if (EMPTY_SET (setB))
    return ! rts_card (setA);
  else if (EMPTY_SET (setA))
    return ! rts_card (setB);
  else if (LO(setA) > HI(setB) || HI(setA) < LO(setB))
    return 0;	/* No overlap, non-empty sets */
  
  {
    int wordsB = NUM_WORDS (setB);
    int wordsA = NUM_WORDS (setA);
      
    if (wordsB == wordsA && LO(setB) == LO(setA))
      return !bcmp (BITV (setA), BITV (setB), NUM_BYTES (setA));
    else
      {
	SET_VECTOR *seta;
	SET_VECTOR *setb;
	int count, equal;
	
	if (LO(setA) < LO(setB))
	  {
	    seta = BITV(setA)+THE_WORD(LO(setB), setA);
	    setb = BITV(setB);
	    if (HI (setA) >= HI (setB))	/* B fits in A range */
	      count = NUM_WORDS (setB);
	    else
	      count = (HI(setA)+BITS_PER_WORD-RLO(setB)) / BITS_PER_WORD;
	  }
	else
	  {
	    seta = BITV(setA);
	    setb = BITV(setB)+THE_WORD(LO(setA), setB);
	    if (HI (setB) >= HI (setA))	/* A fits in B range */
	      count = NUM_WORDS (setA);
	    else
	      count = (HI(setB)+BITS_PER_WORD-RLO(setA)) / BITS_PER_WORD;
	  }

	equal = 1;
	while (count-- && equal)
	  equal = *seta++ == *setb++;

	if (equal)
	  equal = rts_card (setB) == rts_card (setA);

	return equal;
      }
  }
}

/*
 * A <= B (INCLUSION of A in B)
 * 
 * Returns 1 if true, 0 otherwise
 */
int
_p_set_le (setA, setB)
set setA;
set setB;
{
  D(1, Dumpset(setA, 2, "le (a)"));
  D(1, Dumpset(setB, 2, "le (b)"));
  /* First, try the easy cases */
  if (EMPTY_SET (setA))
    return 1;
  else if (EMPTY_SET (setB))
    return ! rts_card (setA);
  else
    {
      int wordsB = NUM_WORDS (setB);
      int wordsA = NUM_WORDS (setA);
      
      SET_VECTOR *b = BITV (setB),
      		 *a = BITV (setA);
      
      if (wordsA <= wordsB && LO(setB) == LO(setA))
	{
	  while (wordsA--)
	    if (*a++ & ~*b++)
	      return 0;
	}
      else {
	/* @@@@@@@@@@@@ finish this */
	_p_error (ABORT, "_p_set_le is under construction, sorry");
      }
      
    }
  return 1;
}

/* Following routines are defined only in Extended Pascal */

/*
 * R = A >< B (SYMMETRIC DIFFERENCE)
 * 
 * Returns a pointer to the resulting set R.
 *
 */

SET_VECTOR  *
_p_set_symdiff (setA, setB, setR)
set setA;
set setB;
set setR;
{
  D(1, Dumpset(setA, 2, "symdif >< (a)"));
  D(1, Dumpset(setB, 2, "symdif >< (b)"));
  /* First, try the easy cases */
  if (EMPTY_SET (setB))
    _p_set_copy (setA, setR);
  else if (EMPTY_SET (setA))
    _p_set_copy (setB, setR);
  else
    {
      int wordsB = NUM_WORDS (setB);
      int wordsR;
      
      SET_VECTOR *b = BITV (setB),
      		 *r = BITV (setR);
      
      _p_set_copy (setA, setR);
      wordsR = NUM_WORDS (setR);

      if (wordsB > wordsR || HI(setB) > HI(setR) || LO(setB) < LO(setR))
	_p_generic (50); /* @@@@ */
      else
	{
	  if (LO(setB) != LO(setR))
	    r += THE_WORD (LO(setB), setR);
	  
	  while (wordsB--)
	    {
	      *r++ ^= *b++;
	    }
	}
    }
  D(1, Dumpset(setR, 2, "symdiff (result)"));
  return BITV(setR);
}

/*
 * CARD(A) (CARD)
 * 
 * Returns a number of members in the set
 *
 */
long
_p_set_card (setA)
set setA;
{
  D(1, Dumpset(setA, 2, "card (a)"));

  return rts_card (setA);
}
